# Form implementation generated from reading ui file 'modify_parameters.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_ModifyParm(object):
    def setupUi(self, ModifyParm):
        ModifyParm.setObjectName("ModifyParm")
        ModifyParm.resize(640, 480)
        self.listWidgetMoleculesNerdss = QtWidgets.QListWidget(parent=ModifyParm)
        self.listWidgetMoleculesNerdss.setGeometry(QtCore.QRect(20, 30, 171, 271))
        self.listWidgetMoleculesNerdss.setObjectName("listWidgetMoleculesNerdss")
        self.listWidgetReactionsNerdss = QtWidgets.QListWidget(parent=ModifyParm)
        self.listWidgetReactionsNerdss.setGeometry(QtCore.QRect(220, 30, 401, 271))
        self.listWidgetReactionsNerdss.setObjectName("listWidgetReactionsNerdss")
        self.label = QtWidgets.QLabel(parent=ModifyParm)
        self.label.setGeometry(QtCore.QRect(70, 10, 81, 17))
        self.label.setObjectName("label")
        self.label_2 = QtWidgets.QLabel(parent=ModifyParm)
        self.label_2.setGeometry(QtCore.QRect(380, 10, 81, 17))
        self.label_2.setObjectName("label_2")
        self.radioButtonSphere = QtWidgets.QRadioButton(parent=ModifyParm)
        self.radioButtonSphere.setGeometry(QtCore.QRect(40, 350, 112, 23))
        self.radioButtonSphere.setObjectName("radioButtonSphere")
        self.radioButtonBox = QtWidgets.QRadioButton(parent=ModifyParm)
        self.radioButtonBox.setGeometry(QtCore.QRect(40, 390, 112, 23))
        self.radioButtonBox.setObjectName("radioButtonBox")
        self.pushButtonOkModifyParm = QtWidgets.QPushButton(parent=ModifyParm)
        self.pushButtonOkModifyParm.setGeometry(QtCore.QRect(190, 440, 89, 25))
        self.pushButtonOkModifyParm.setObjectName("pushButtonOkModifyParm")
        self.pushButtonCancelModifyParm = QtWidgets.QPushButton(parent=ModifyParm)
        self.pushButtonCancelModifyParm.setGeometry(QtCore.QRect(350, 440, 89, 25))
        self.pushButtonCancelModifyParm.setObjectName("pushButtonCancelModifyParm")
        self.pushButtonSetGeometry = QtWidgets.QPushButton(parent=ModifyParm)
        self.pushButtonSetGeometry.setGeometry(QtCore.QRect(130, 370, 131, 25))
        self.pushButtonSetGeometry.setObjectName("pushButtonSetGeometry")
        self.pushButtonModifyMol = QtWidgets.QPushButton(parent=ModifyParm)
        self.pushButtonModifyMol.setGeometry(QtCore.QRect(20, 310, 171, 25))
        self.pushButtonModifyMol.setObjectName("pushButtonModifyMol")
        self.pushButtonModifyReaction = QtWidgets.QPushButton(parent=ModifyParm)
        self.pushButtonModifyReaction.setGeometry(QtCore.QRect(330, 310, 171, 25))
        self.pushButtonModifyReaction.setObjectName("pushButtonModifyReaction")
        self.label_3 = QtWidgets.QLabel(parent=ModifyParm)
        self.label_3.setGeometry(QtCore.QRect(340, 380, 161, 17))
        self.label_3.setObjectName("label_3")
        self.lineEditTime = QtWidgets.QLineEdit(parent=ModifyParm)
        self.lineEditTime.setGeometry(QtCore.QRect(490, 380, 113, 25))
        self.lineEditTime.setObjectName("lineEditTime")

        self.retranslateUi(ModifyParm)
        QtCore.QMetaObject.connectSlotsByName(ModifyParm)

    def retranslateUi(self, ModifyParm):
        _translate = QtCore.QCoreApplication.translate
        ModifyParm.setWindowTitle(_translate("ModifyParm", "Modify Parameters"))
        self.label.setText(_translate("ModifyParm", "Molecules"))
        self.label_2.setText(_translate("ModifyParm", "Reactions"))
        self.radioButtonSphere.setText(_translate("ModifyParm", "Sphere"))
        self.radioButtonBox.setText(_translate("ModifyParm", "Box"))
        self.pushButtonOkModifyParm.setText(_translate("ModifyParm", "Ok"))
        self.pushButtonCancelModifyParm.setText(_translate("ModifyParm", "Cancel"))
        self.pushButtonSetGeometry.setText(_translate("ModifyParm", "Set geometry"))
        self.pushButtonModifyMol.setText(_translate("ModifyParm", "Modify molecule"))
        self.pushButtonModifyReaction.setText(_translate("ModifyParm", "Modify reaction"))
        self.label_3.setText(_translate("ModifyParm", "Simulation length (s)"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ModifyParm = QtWidgets.QDialog()
    ui = Ui_ModifyParm()
    ui.setupUi(ModifyParm)
    ModifyParm.show()
    sys.exit(app.exec())
